
static void
print_something ( void  )
{
	printf ("\n" "Hello World (from the virtual machine).\n\n\n" );
}

static void
vm_create_replica ( struct vm *vm, unsigned long guest_image_start, unsigned long guest_image_size, unsigned long vm_pmem_size )
{
	struct vmcb *vmcb;

	vmcb = alloc_vmcb ( );
	vm->vmcb = vmcb;

	vmcb->guest_asid  = 1;
        vmcb->tlb_control = 1;	
	vmcb->general2_intercepts = INTRCPT_VMRUN;
	vmcb->iopm_base_pa  = create_intercept_table ( 12 << 10 /* 12 KB */ );
	vmcb->msrpm_base_pa = create_intercept_table ( 8 << 10  /*  8 KB */ );

	vmcb->np_enable = 1;
	printf ( "Nested paging is enabled.\n" );

	{
		u32 eax, edx;
		rdmsr ( MSR_PAT, eax, edx );
		vmcb->g_pat = ( ( ( (u64) edx ) << 32 ) | eax );
	}

	/* [Note] vm_pmem_start is a virtual base address of the physical memory of the virtual machine.  */
	const unsigned long vm_pmem_start = alloc_vm_pmem ( vm_pmem_size );
	vmcb->h_cr3 = create_vm_pmem_mapping_table ( vm_pmem_start, vm_pmem_size );
	

	/* Copy the guest image. */
	memmove ( ( void * ) vm_pmem_start, VIRT ( 0 ), vm_pmem_size );  

	vm->vmcb->rip = ( unsigned long )( &print_something );

	{
		unsigned long rsp;
		__asm__ __volatile__ ( "movq %%rsp, %0" : "=r" ( rsp ) );
		vm->vmcb->rsp = rsp;
	}

	{
		vmcb->cs.sel   = VMM_CS64;
//		vmcb->cs.attrs.bytes = 0x089b; // 1000 1001 1011
		vmcb->cs.attrs.fields.type = 0xa;
		vmcb->cs.attrs.fields.l = 1;
		vmcb->cs.attrs.fields.g = 1;
		vmcb->cs.limit = 0xffff;
		vmcb->cs.base  =0;
	}

	// type = 1011 : Code-segment, R(eadable), A(ccessed)
	// S(ystem/user), P(resent), G(ranularity)

	int i;
	struct seg_selector *segregs [] = { &vmcb->ss, &vmcb->ds, &vmcb->es, &vmcb->fs, &vmcb->gs, NULL };
	for ( i = 0; segregs [ i ] != NULL; i++ ) {
		struct seg_selector *x = segregs [ i ];
		x->sel   = VMM_DS64;
		x->attrs.bytes = 0x0893; // 1000 1001 0011 
		x->limit = 0xffff;
		x->base  = 0;

		// type = 0011 : Data-segment, W(ritable), A(ccessed)
		// S(ystem/user), P(resent), G(ranularity)
	}

	{
		u64 dtr;
		__asm__ __volatile__ ( "sgdt %0":"=m" ( dtr ) );
		vmcb->gdtr.sel   = 0;
		vmcb->gdtr.attrs.bytes = 0;
		vmcb->gdtr.limit = dtr & 0xff;
		vmcb->gdtr.base  = dtr >> 16;
	}

	{
		unsigned long rflags;
		__asm__ __volatile__( "pushfq; popq %0" :"=m" ( rflags ) );
		vmcb->rflags = rflags & 2;
	}

	vmcb->cr0 = read_cr0 ();
	vmcb->cr3 = read_cr3 (); /* guest */
	vmcb->cr4 = read_cr4 ();

	{
		u32 eax, edx;
		rdmsr ( MSR_EFER, eax, edx );
		vmcb->efer = ( ( ( (u64) edx) << 32 ) | eax );
	}

	printf ( "New virtual machine created.\n" ); 
}

void
vm_create ( struct vm *vm, unsigned long guest_image_start, unsigned long guest_image_size, unsigned long vm_pmem_size )
{
	vm_create_replica ( vm, guest_image_start, guest_image_size, vm_pmem_size );
}
